import datetime
import pytz
import re
import sys

from django.conf import settings
from django.contrib.auth import authenticate
from django.contrib.auth import get_user_model
from django.contrib.auth.tokens import default_token_generator
from django.core import mail as Mail
from django.core.exceptions import ValidationError as ModelValidationError
from django.shortcuts import resolve_url
from django.shortcuts import reverse
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django.test import TestCase
from rest_framework.test import APIClient

from .utils import strings

if sys.version_info[0] < 3:
    # python 2
    import mock
else:
    # python 3
    from unittest import mock


User = get_user_model()
timezone = pytz.timezone(settings.TIME_ZONE)


class TestAuthViewBase(TestCase):

    url = ''
    payload = {}
    client = APIClient()
    username = strings[0]
    password = strings[0]

    def _setup_user(self):
        self.user = User.objects.create_user(username=self.username, password=self.password)


class TestLoginView(TestAuthViewBase):

    url = reverse('auth-login')

    def setUp(self):
        self.payload = dict(username=self.username, password=self.password)

    def test_get(self):
        response = self.client.get(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 200)
        self._setup_user()
        self.client.force_login(self.user)
        response = self.client.get(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 405)

    def test_post(self):
        self._setup_user()
        self.assertNotIn('_auth_user_id', self.client.session)
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 302)
        self.assertEquals(response._headers['location'][1], resolve_url(settings.LOGIN_REDIRECT_URL))
        self.assertIn('_auth_user_id', self.client.session)

    def test_invalid_username(self):
        self._setup_user()
        self.payload['username'] += strings[0]
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 400)
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_invalid_password(self):
        self._setup_user()
        self.payload['password'] += strings[0]
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 400)
        self.assertNotIn('_auth_user_id', self.client.session)

    def test_invalid_form(self):
        self.payload['password'] = ''
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 400)


class TestLogoutView(TestAuthViewBase):

    url = reverse('auth-logout')

    def test_get(self):
        response = self.client.get(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 405)
        self._setup_user()
        self.client.force_login(self.user)
        response = self.client.get(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 200)

    def test_post(self):
        self._setup_user()
        self.client.force_login(self.user)
        self.assertIn('_auth_user_id', self.client.session)
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 200)
        self.assertNotIn('_auth_user_id', self.client.session)


class TestChangePasswordView(TestAuthViewBase):

    url = reverse('auth-change-password')
    valid_password = 'abcdefghijklmnopqrstuvwxyz'

    def setUp(self):
        self.payload = dict(password=self.valid_password, repeat_password=self.valid_password)

    def test_get(self):
        response = self.client.get(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 405)
        self._setup_user()
        self.client.force_login(self.user)
        response = self.client.get(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 200)

    @mock.patch('drf_advanced_auth.serializers.validate_password')
    def test_post(self, mock_validate_password):
        self._setup_user()
        self.assertIsNotNone(authenticate(username=self.username, password=self.password))
        self.client.force_login(self.user)
        response = self.client.post(self.url, self.payload, format='json')
        mock_validate_password.assert_called_once()
        self.assertEquals(response.status_code, 200)
        self.assertIsNone(authenticate(username=self.username, password=self.password))
        self.assertIsNotNone(authenticate(username=self.username, password=self.valid_password))

    @mock.patch('drf_advanced_auth.serializers.validate_password')
    def test_invalid_password(self, mock_validate_password):
        self._setup_user()
        self.client.force_login(self.user)
        self.payload = dict(password=strings[0], repeat_password=strings[0])
        mock_validate_password.side_effect = ModelValidationError([])
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 400)
        self.assertIsNotNone(authenticate(username=self.username, password=self.password))

    def test_multifield_password_mismatch(self):
        self._setup_user()
        self.client.force_login(self.user)
        self.payload['repeat_password'] += strings[0]
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 400)
        self.assertIsNotNone(authenticate(username=self.username, password=self.password))


class TestAuthViewResetPasswordBase(TestAuthViewBase):

    valid_email_address = 'test@test.com'
    valid_wrong_email_address = 'tset@test.com'
    invalid_email_address = 'test@test'
    valid_password = 'abcdefghijklmnopqrstuvwxyz'

    def _setup_user(self, include_token=False):
        super()._setup_user()
        self.user.email = self.valid_email_address
        self.user.last_login = timezone.localize(datetime.datetime.now())
        self.user.save()
        if include_token:
            self.client.force_login(self.user)
            self.client.logout()
            self.user = User.objects.get(pk=self.user.pk)
            self.payload = dict(
                uidb64=urlsafe_base64_encode(force_bytes(self.user.pk)),
                token=default_token_generator.make_token(self.user),
            )


class TestAuthViewResetPasswordRequest(TestAuthViewResetPasswordBase):

    url = reverse('auth-reset-password-request')
    reset_link_regex = '(https?://[^/]+{0})'.format(
        reverse('password_reset_confirm', kwargs=dict(uidb64='.*', token='.*'))
    )

    def setUp(self):
        self.payload = dict(email=self.valid_email_address)

    def test_get(self):
        response = self.client.get(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 200)
        self._setup_user()
        self.client.force_login(self.user)
        response = self.client.get(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 405)

    def test_post(self):
        self._setup_user()
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 200)
        self.assertEquals(len(Mail.outbox), 1)
        mail = Mail.outbox[0]
        self.assertEquals(list(mail.to), [self.valid_email_address])
        self.assertEquals(mail.subject, 'Password reset on testserver')
        self.assertIsNotNone(re.search(self.reset_link_regex, mail.body))

    def test_empty_email(self):
        self._setup_user()
        self.payload['email'] = self.valid_wrong_email_address
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 200)
        self.assertEquals(len(Mail.outbox), 0)

    def test_invalid_email(self):
        self._setup_user()
        self.payload['email'] = self.invalid_email_address
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 400)


class TestAuthViewResetPasswordComplete(TestAuthViewResetPasswordBase):

    url = reverse('auth-reset-password-complete')

    def _setup_user(self):
        super(TestAuthViewResetPasswordComplete, self)._setup_user(include_token=True)
        self.payload.update(dict(
            password=self.valid_password,
            repeat_password=self.valid_password,
        ))

    def test_get(self):
        response = self.client.get(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 200)
        self._setup_user()
        self.client.force_login(self.user)
        response = self.client.get(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 405)

    def test_post(self):
        self._setup_user()
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 302)
        self.assertEquals(response._headers['location'][1], resolve_url(settings.LOGIN_REDIRECT_URL))
        self.assertIn('_auth_user_id', self.client.session)

    def test_invalid_token(self):
        self._setup_user()
        self.payload['token'] = strings[0]
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 400)

    def test_invalid_uidb64(self):
        self._setup_user()
        self.payload['uidb64'] = strings[0]
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 400)

    @mock.patch('drf_advanced_auth.serializers.validate_password')
    def test_invalid_password(self, mock_validate_password):
        self._setup_user()
        mock_validate_password.side_effect = ModelValidationError([])
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 400)
        assert authenticate(username=self.user.username, password=self.password) is not None

    def test_multifield_password_mismatch(self):
        self._setup_user()
        self.payload['repeat_password'] += strings[0]
        response = self.client.post(self.url, self.payload, format='json')
        self.assertEquals(response.status_code, 400)
        self.assertIsNotNone(authenticate(username=self.user.username, password=self.password))



from django.conf import settings
from django.contrib.auth import authenticate
from django.contrib.auth import login
from django.contrib.auth import logout
from django.http import HttpResponseRedirect
from django.shortcuts import resolve_url
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet

from . import serializers
from . import utils


_success_response = dict(status='OK')
_user_logged_in_error = dict(error='user logged in')
_user_logged_out_error = dict(error='user logged out')
_authentication_failure_error = dict(error='authentication error')
_invalid_token_error = dict(error='invalid token')


class AuthViewSet(GenericViewSet):

    permission_classes = []

    serializer_for_action = dict(
        login=serializers.LoginSerializer,
        logout=serializers.NullSerializer,
        change_password=serializers.ChangePasswordSerializer,
        reset_password_request=serializers.ResetPasswordRequestSerializer,
        reset_password_complete=serializers.ResetPasswordCompleteSerializer,
    )

    def get_serializer_class(self):
        return self.serializer_for_action.get(self.action, serializers.NullSerializer)

    @action(methods=['get', 'post'], detail=False)
    def login(self, request):
        if request.user.is_authenticated:
            return Response(_user_logged_in_error, status=status.HTTP_405_METHOD_NOT_ALLOWED)
        if request.method == 'GET':
            serializer = self.get_serializer()
            return Response(serializer.data)
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = authenticate(username=serializer.data['username'], password=serializer.data['password'])
        if user is None:
            return Response(_authentication_failure_error, status=status.HTTP_400_BAD_REQUEST)
        login(request, user)
        return HttpResponseRedirect(resolve_url(settings.LOGIN_REDIRECT_URL))

    @action(methods=['get', 'post'], detail=False)
    def logout(self, request):
        if not request.user.is_authenticated:
            return Response(_user_logged_out_error, status=status.HTTP_405_METHOD_NOT_ALLOWED)
        if request.method == 'GET':
            serializer = self.get_serializer()
            return Response(serializer.data)
        logout(request)
        return Response(_success_response)

    @action(methods=['get', 'post'], detail=False)
    def change_password(self, request):
        if not request.user.is_authenticated:
            return Response(_user_logged_out_error, status=status.HTTP_405_METHOD_NOT_ALLOWED)
        if request.method == 'GET':
            serializer = self.get_serializer()
            return Response(serializer.data)
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(_success_response)

    @action(methods=['get', 'post'], detail=False)
    def reset_password_request(self, request):
        if request.user.is_authenticated:
            return Response(_user_logged_in_error, status=status.HTTP_405_METHOD_NOT_ALLOWED)
        if request.method == 'GET':
            serializer = self.get_serializer()
            return Response(serializer.data)
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        utils.send_mail_to_matching_users(request, serializer.data['email'])
        return Response(_success_response)

    @action(methods=['get', 'post'], detail=False)
    def reset_password_complete(self, request):
        if request.user.is_authenticated:
            return Response(_user_logged_in_error, status=status.HTTP_405_METHOD_NOT_ALLOWED)
        if request.method == 'GET':
            serializer = self.get_serializer()
            return Response(serializer.data)
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        login(request, serializer.user, backend=settings.AUTHENTICATION_BACKENDS[0])
        return HttpResponseRedirect(resolve_url(settings.LOGIN_REDIRECT_URL))



import django


DEBUG = True
USE_TZ = True

SECRET_KEY = 'abcdefghijklmnopqrstuvwxyz'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'db.sqlite',
    }
}

ROOT_URLCONF = 'sample_app.urls'

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'rest_framework',
    'sample_app',
)

SITE_ID = 1

_middleware_list = (
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
)

if django.VERSION >= (1, 10):
    MIDDLEWARE = _middleware_list
else:
    MIDDLEWARE_CLASSES = _middleware_list

TEMPLATES = (
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': (),
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
)

AUTH_USER_MODEL = 'sample_app.CustomUser'

APPEND_SLASH = True
STATIC_URL = '/static/'

REST_FRAMEWORK = {
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'sample_app.authentication.CsrfExemptSessionAuthentication',
    ),
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.AllowAny',
    )
}


LOGIN_REDIRECT_URL = 'index'


DRF_ADVANCED_AUTH = {
    'AUTH_TOKEN_LIFETIME_MINUTES': 15,
}



from django.conf.urls import include
from django.conf.urls import url
from django.views import View
from django.views.generic import TemplateView


class FakeView(View):
    pass


urlpatterns = [
    url(r'^auth/', include('drf_advanced_auth.urls')),
    url(r'^password-reset/update/(?P<uidb64>.*)/(?P<token>.*)', FakeView.as_view(), name='password_reset_confirm'),
    url(r'^$', TemplateView.as_view(template_name='index.html'), name='index'),
]



from django.contrib.auth import get_user_model
from django.test import TestCase

from sample_app.models import CustomUser


class TestAssumptions(TestCase):

    def test_user_is_right_user(self):
        self.assertTrue(issubclass(CustomUser, get_user_model()))



from django.contrib.auth.password_validation import validate_password
from django.core.exceptions import ValidationError as ModelValidationError
from rest_framework import serializers

from . import utils


class PasswordField(serializers.CharField):

    def __init__(self, *args, **kwargs):
        super(PasswordField, self).__init__(*args, **kwargs)
        style = self.style
        if 'input_type' in style:
            raise Exception('do not override field_type for password fields')
        self.style['input_type'] = 'password'


NullSerializer = serializers.Serializer


class ChangePasswordSerializer(serializers.Serializer):

    password = PasswordField()
    repeat_password = PasswordField()

    def validate_password(self, value):
        request = self.context.get('request', None)
        user = request.user if request and hasattr(request, 'user') else None
        try:
            validate_password(value, user=user)
        except ModelValidationError as e:
            raise serializers.ValidationError(e)
        return value

    def validate(self, data):
        if data['repeat_password'] != data['password']:
            raise serializers.ValidationError('passwords mismatch')
        return data

    def save(self):
        user = self.context['request'].user
        user.set_password(self.data['password'])
        user.save()


class LoginSerializer(serializers.Serializer):

    username = serializers.CharField()
    password = PasswordField()


class ResetPasswordCompleteSerializer(ChangePasswordSerializer):

    uidb64 = serializers.CharField()
    token = serializers.CharField()

    def validate(self, data):
        data = super(ResetPasswordCompleteSerializer, self).validate(data)
        try:
            self.user = utils.check_token(data['uidb64'], data['token'])
        except Exception as e:
            raise serializers.ValidationError(e)
        return data

    def save(self):
        user = self.user
        user.set_password(self.data['password'])
        user.save()


class ResetPasswordRequestSerializer(serializers.Serializer):

    email = serializers.EmailField()



from rest_framework.authentication import SessionAuthentication


class CsrfExemptSessionAuthentication(SessionAuthentication):

    def enforce_csrf(self, request):
        return  # disable all CSRF protections locally - DO NOT DO THIS IN PRODUCTION



import os

from setuptools import find_packages
from setuptools import setup


try:
    with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
        README = readme.read()
except Exception:
    README = '<failed to open README.md>'


os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


install_dependencies = (
    'Django>=1.11',
)

test_dependencies = (
    'mock',
    'djangorestframework',
)


setup(
    name='drf_advanced_auth',
    version='0.1.1',
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',
    description='Full authentication and credential management for Django REST Framework',
    long_description=README,
    long_description_content_type='text/markdown',
    author='jeffrey k eliasen',
    author_email='jeff+drf-advanced-auth@jke.net',
    url='https://github.com/seawolf42/drf-advanced-auth',
    zip_safe=False,
    keywords='drf-advanced-auth',
    classifiers=[
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    install_requires=install_dependencies,
    tests_require=install_dependencies + test_dependencies,
    test_suite='runtests.run_tests',
)



import datetime

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.tokens import default_token_generator
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import send_mail
from django.template import loader
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_decode
from django.utils.http import urlsafe_base64_encode

from . import conf


_token_lifetime = datetime.timedelta(days=conf.AUTH_TOKEN_LIFETIME_MINUTES)


def check_token(uidb64, token):
    user = get_user_model().objects.get(pk=urlsafe_base64_decode(uidb64).decode())
    if not default_token_generator.check_token(user, token):
        raise Exception
    return user


def send_mail_to_matching_users(request, email):
    site = get_current_site(request)
    for user in (
        u for u in get_user_model()._default_manager.filter(email__iexact=email, is_active=True)
        if u.has_usable_password()
    ):
        context = dict(
            email=email,
            domain=site.domain,
            site_name=site.name,
            uid=urlsafe_base64_encode(force_bytes(user.pk)),
            user=user,
            token=default_token_generator.make_token(user),
            protocol=request.scheme,
        )
        subject = ''.join(
            loader.render_to_string('registration/password_reset_subject.txt', context).splitlines()
        )
        body = loader.render_to_string('registration/password_reset_email.html', context)
        send_mail(subject, body, settings.SERVER_EMAIL, [email])
